import java.util.*;

public class Compiler {
    static interface Expression {
        Real evaluate(Hashtable bindings);
    }

    static interface Declaration {
        void declare(Hashtable bindings);
    }

    static class ExpressionList implements Expression {
        Expression[] exprs;

        public Real evaluate(Hashtable bindings) {
            Real val = exprs[0].evaluate(bindings);
            for (int i = 1; i < exprs.length; i++) {
                val = exprs[i].evaluate(bindings);
            }
            return val;
        }

        ExpressionList(Expression[] exprs) {
            this.exprs = exprs;
        }
    }

    static class AssignmentExpression implements Expression, Declaration {
        String     name;
        Expression expr;

        public void declare(Hashtable bindings) {
            bindings.put(name, expr);
        }

        public Real evaluate(Hashtable bindings) {
            declare(bindings);
            return expr.evaluate(bindings);
        }
    }

    static class FunctionDefExpression implements Expression, Declaration {
        String             name;
        FunctionExpression func;

        public void declare(Hashtable bindings) {
            bindings.put(name, func);
        }

        public Real evaluate(Hashtable bindings) {
            declare(bindings);
            return null;
        }
    }

    static class ConditionalExpression implements Expression {
        Expression cond;
        Expression ifTrue;
        Expression ifFalse;

        public Real evaluate(Hashtable bindings) {
            return cond.evaluate(bindings).equalTo(Real.ZERO)
                ? ifFalse.evaluate(bindings)
                : ifTrue.evaluate(bindings);
        }
    }

    abstract static class UnaryExpression implements Expression {
        Expression operand;
    }

    static class NegExpression extends UnaryExpression {
        public Real evaluate(Hashtable bindings) {
            return operand.evaluate(bindings).Neg();
        }

        NegExpression(Expression operand) {
            this.operand = operand;
        }
    }

    static class VariableExpression implements Expression {
        String name;

        public Real evaluate(Hashtable bindings) {
            Expression expr = (Expression)bindings.get(name);
            if (expr == null) {
                throw new CompileError(Locale.RES_VARIABLE + name + Locale.RES_NOT_DEFINED);
            }
            return expr.evaluate(bindings);
        }

        VariableExpression(String name) {
            this.name = name;
        }
    }

    static class FunctionExpression implements Expression {
        Real[]    arguments;
        Expression body;

        public Real evaluate(Hashtable bindings) {
            return body.evaluate(bindings);
        }
    }

    static class ParameterExpression implements Expression {
        FunctionExpression function;
        int                index;

        public Real evaluate(Hashtable bindings) {
            return function.arguments[index];
        }

        public ParameterExpression(int index, FunctionExpression function) {
            this.index = index;
            this.function = function;
        }
    }

    static class PlotGraphExpression implements Expression {
        Expression[] coords;
        Point[]      points;

        PlotGraphExpression(Expression[] coords) { 
            this.coords = coords;
        }

        public Real evaluate(Hashtable bindings) {
            int n = coords.length/2;
            Point[] points = new Point[n];
            for (int i = 0; i < n; i++) { 
                Expression x = coords[i*2];
                Expression y = coords[i*2+1];
                points[i] = new Point(x.evaluate(bindings), y.evaluate(bindings));
            }
            this.points = points;
            return null;
        }
    }
                
        
    static class PlotExpression implements Expression {
        Expression from;
        Expression till;
        Expression step;
        String     func;

        FunctionExpression f;

        public Real evaluate(Hashtable bindings) {
            Expression expr = (Expression)bindings.get(func);
            if (expr == null) {
                Integer kind = (Integer)builtin.get(func);
                if (kind != null) {
                    // builtin function
                    f = new FunctionExpression();
                    f.arguments = new Real[1];
                    f.body = new BuiltinFunction(kind.intValue(), new ParameterExpression(0, f));
                }  else {
                    throw new CompileError(Locale.RES_FUNCTION + func + Locale.RES_NOT_DEFINED);
                }
            } else {
                if (!(expr instanceof FunctionExpression)) {
                    throw new CompileError(Locale.RES_LAST_PLOT_PARAM);
                }
                f = (FunctionExpression)expr;
                if (f.arguments.length != 1) {
                    throw new CompileError(Locale.RES_LAST_PLOT_PARAM_ONEARG);
                }
            }
            return null;
        }
    }

    static class FunctionCallExpression implements Expression {
        String name;
        Expression[] parameters;

        public Real evaluate(Hashtable bindings) {
            Expression expr = (Expression)bindings.get(name);
            if (expr == null) {
                throw new CompileError(Locale.RES_FUNCTION + name + Locale.RES_NOT_DEFINED);
            }
            if (!(expr instanceof FunctionExpression)) {
                throw new CompileError(Locale.RES_NAME + name + Locale.RES_NOT_FUNCTION_NAME);
            }
            FunctionExpression func = (FunctionExpression)expr;
            Real[] arguments = new Real[parameters.length];
            for (int i = 0; i < parameters.length; i++) {
                arguments[i] = parameters[i].evaluate(bindings);
            }
            Real[] saveArguments = func.arguments;
            func.arguments = arguments;
            Real result = func.evaluate(bindings);
            func.arguments = saveArguments;
            return result;
        }
    }


    static class LiteralExpression implements Expression {
        Real value;

        public Real evaluate(Hashtable bindings) {
            return value;
        }

        LiteralExpression(Real value) {
            this.value = value;
        }
    }

    static final int SIN = 0;
    static final int COS = 1;
    static final int TAN = 2;
    static final int ASIN = 3;
    static final int ACOS = 4;
    static final int ATAN = 5;
    static final int EXP = 6;
    static final int LOG = 7;
    static final int ABS = 8;
    static final int CEIL = 9;
    static final int FLOOR  = 10;
    static final int INT = 11;
    static final int FRAC  = 12;
    static final int SQR = 13;
    static final int LN = 14;

    static Hashtable builtin;
    static {
        builtin = new Hashtable();
        builtin.put("sin", new Integer(SIN));
        builtin.put("cos", new Integer(COS));
        builtin.put("tan", new Integer(TAN));
        builtin.put("asin", new Integer(ASIN));
        builtin.put("acos", new Integer(ACOS));
        builtin.put("atan", new Integer(ATAN));
        builtin.put("exp", new Integer(EXP));
        builtin.put("sqr", new Integer(SQR));
        builtin.put("log", new Integer(LOG));
        builtin.put("ln", new Integer(LN));
        builtin.put("ceil", new Integer(CEIL));
        builtin.put("floor", new Integer(FLOOR));
    }

    static class BuiltinFunction implements Expression {
        Expression operand;
        int kind;

        public Real evaluate(Hashtable bindings) {
            Real x = new Real(operand.evaluate(bindings));
            switch (kind) {
            case SIN:
                x.sin();
                break;
            case COS:
                x.cos();
                break;
            case TAN:
                x.tan();
                break;
            case ASIN:
                x.asin();
                break;
            case ACOS:
                x.acos();
                break;
            case ATAN:
                x.atan();
                break;
            case EXP:
                x.exp();
                break;
            case LN:
                x.ln();
                break;
            case LOG:
                x.log10();
                break;
            case ABS:
                x.abs();
                break;
            case CEIL:
                x.ceil();
                break;
            case FLOOR:
                x.floor();
            case SQR:
                x.sqrt();
                break;
            case INT:
                x.trunc();
                break;
            case FRAC:
                x.frac();
                break;
            default:
                throw new CompileError(Locale.RES_INVALID_FUNCTION);
            }
            return x;
        }

        BuiltinFunction(int kind, Expression operand) {
            this.kind = kind;
            this.operand = operand;
        }
    }

    static abstract class BinaryExpression implements Expression {
        Expression left;
        Expression right;
    }

    static class LessExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).lessThan(right.evaluate(bindings))
                ? Real.ONE : Real.ZERO;
        }
    }

    static class GreatExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).greaterThan(right.evaluate(bindings))
                ? Real.ONE : Real.ZERO;
        }
    }

    static class RangeExpression implements Expression {
        Expression min;
        Expression expr;
        Expression max;

        public Real evaluate(Hashtable bindings) {
            Real val = expr.evaluate(bindings);
            return val.greaterThan(min.evaluate(bindings)) && val.lessThan(max.evaluate(bindings))
                ? Real.ONE : Real.ZERO;
        }
    }

    static class MulExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).Mul(right.evaluate(bindings));
        }
    }

    static class AddExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).Add(right.evaluate(bindings));
        }
    }

    static class SubExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).Sub(right.evaluate(bindings));
        }
    }

    static class DivExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).Div(right.evaluate(bindings));
        }
    }

    static class PowerExpression extends BinaryExpression {
        public Real evaluate(Hashtable bindings) {
            return left.evaluate(bindings).Pow(right.evaluate(bindings));
        }
    }

    public Compiler(String str) {
        this.str = str;
    }

    public Expression compile() {
        Expression expr = compileExpression();
        if (tkn != -1) {
            throw new CompileError(Locale.RES_UNEXPECTED_TOKEN + (char)tkn + "'");
        }
        return expr;
    }

    private Expression compileExpression() {
        Expression expr = compileAssignmentExpression();
        if (tkn == ',') {
            Vector exprList = new Vector();
            exprList.addElement(expr);
            do {
                exprList.addElement(compileAssignmentExpression());
            } while (tkn == ',');

            Expression[] exprs = new Expression[exprList.size()];
            for (int i = 0; i < exprs.length; i++) {
                exprs[i] = (Expression)exprList.elementAt(i);
            }
            expr = new ExpressionList(exprs);
        }
        return expr;
    }

    private Expression compileAssignmentExpression() {
        Expression lvalue = compileCondExpression();
        if (tkn == '=') {
            if (lvalue instanceof VariableExpression) {
                AssignmentExpression asg = new AssignmentExpression();
                asg.name = ((VariableExpression)lvalue).name;
                asg.expr = compileAssignmentExpression();
                return asg;
             } else if (lvalue instanceof FunctionCallExpression) {
                FunctionCallExpression call = (FunctionCallExpression)lvalue;
                FunctionExpression outer = function;
                String[] outerParameters = parameters;
                parameters = new String[call.parameters.length];
                for (int i = 0; i < parameters.length; i++) {
                    if (call.parameters[i] instanceof VariableExpression) {
                        parameters[i] = ((VariableExpression)call.parameters[i]).name;
                    } else {
                        throw new CompileError(Locale.RES_INVALID_FUNC_PARAMETER);
                    }
                }
                function = new FunctionExpression();
                function.arguments = new Real[parameters.length];
                function.body = compileAssignmentExpression();
                FunctionDefExpression defun = new FunctionDefExpression();
                defun.name = call.name;
                defun.func = function;
                function = outer;
                parameters = outerParameters;
                return defun;
            } else {
                throw new CompileError(Locale.RES_ASSIGN_ONLY_VAR_OR_FUNC);
            }
        }
        return lvalue;
    }

    private Expression compileCondExpression() {
        Expression left = compileCmpExpression();
        while (tkn == '?') {
            ConditionalExpression cond = new ConditionalExpression();
            cond.cond = left;
            cond.ifTrue = compileCmpExpression();
            if (tkn != ':') {
                throw new CompileError(Locale.RES_MISSING_COLON);
            }
            cond.ifFalse = compileCmpExpression();
            left = cond;
        }
        return left;
    }

    private Expression compileCmpExpression() {
        Expression left = compileAddExpression();
        int cop = tkn;
        if (cop == '>' || cop == '<') {
            Expression right = compileAddExpression();
            if (cop == tkn) {
                RangeExpression range = new RangeExpression();
                if (cop == '<') {
                    range.min = left;
                    range.expr = right;
                    range.max = compileAddExpression();
                } else {
                    range.max = left;
                    range.expr = right;
                    range.min = compileAddExpression();
                }
                return range;
            } else {
                BinaryExpression bin = (cop == '<')
                    ? (BinaryExpression)new LessExpression() : (BinaryExpression)new GreatExpression();
                bin.left = left;
                bin.right = right;
                return bin;
            }
        }
        return left;
    }


    private Expression compileAddExpression() {
        Expression left = compileMulExpression();
        int cop = tkn;
        while (cop == '+' || cop == '-') {
            Expression right =  compileMulExpression();
            BinaryExpression bin = (cop == '+')
                ? (BinaryExpression)new AddExpression() : (BinaryExpression)new SubExpression();
            bin.left = left;
            bin.right = right;
            left = bin;
            cop = tkn;
        }
        return left;
    }

    private Expression compileMulExpression() {
        Expression left = compilePowerExpression();
        int cop = tkn;
        while (cop == '*' || cop == '/') {
            Expression right =  compilePowerExpression();
            BinaryExpression bin = (cop == '*')
                ? (BinaryExpression)new MulExpression() : (BinaryExpression)new DivExpression();
            bin.left = left;
            bin.right = right;
            left = bin;
            cop = tkn;
        }
        return left;
    }

    private Expression compilePowerExpression() {
        Expression left = compileUnaryExpression();
        if (tkn == '^') {
            Expression right = compilePowerExpression();
            PowerExpression pow = new PowerExpression();
            pow.left = left;
            pow.right = right;
            return pow;
        }
        return left;
    }

    private Expression compileUnaryExpression() {
        Expression expr;
        switch (scan()) {
        case '+':
            return compileUnaryExpression();
        case '-':
            return new NegExpression(compileUnaryExpression());
        case '0':
            expr = new LiteralExpression(value);
            break;
        case '(':
            expr = compileExpression();
            if (tkn != ')') {
                throw new CompileError(Locale.RES_MISSING_BRACKET);
            }
            break;
        case '[':
            expr = compileExpression();
            if (tkn != ']') {
                throw new CompileError(Locale.RES_MISSING_SQBRACKET);
            }
            expr = new BuiltinFunction(INT, expr);
            break;
        case '{':
            expr = compileExpression();
            if (tkn != '}') {
                throw new CompileError(Locale.RES_MISSING_THRDBRACKET);
            }
            expr = new BuiltinFunction(FRAC, expr);
            break;
        case '|':
            expr = compileExpression();
            if (tkn != '|') {
                throw new CompileError(Locale.RES_MISSING_VERBAR);
            }
            expr = new BuiltinFunction(ABS, expr);
            break;
        case 'a':
            if (name.equals("pi")) {
                expr = new LiteralExpression(Real.PI);
            } else if (name.equals("e")) {
                expr = new LiteralExpression(Real.E);
            } else {
                if (function != null) {
                    for (int i = 0; i < parameters.length; i++) {
                        if (parameters[i].equals(name)) {
                            tkn = scan();
                            return new ParameterExpression(i, function);
                        }
                    }
                }
                String ident = name;
                tkn = scan();
                if (tkn == '(') {
                    expr = compileExpression();
                    if (tkn != ')') {
                        throw new CompileError(Locale.RES_MISSING_BRACKET);
                    }
                    Expression[] params;
                    if (expr instanceof ExpressionList) {
                        params = ((ExpressionList)expr).exprs;
                    } else {
                        params = new Expression[1];
                        params[0] = expr;
                    }
                    if (ident.equals("plot")) {
                        if (params.length < 3) {
                            throw new CompileError(Locale.RES_PLOT_3OR4_PARAMS);
                        }
                        PlotExpression plot = new PlotExpression();
                        plot.from = params[0];
                        plot.till = params[1];
                        Expression func;
                        if (params.length == 4) {
                            plot.step = params[2];
                            func = params[3];
                        } else {
                            func = params[2];
                            plot.step = null;
                        }
                        if (!(func instanceof VariableExpression)) {
                            throw new CompileError(Locale.RES_LAST_PLOT_PARAM);
                        }
                        plot.func = ((VariableExpression)func).name;
                        expr = plot;
                    } else {
                        Integer kind = (Integer)builtin.get(ident);
                        if (kind != null) {
                            if (params.length != 1) {
                                throw new CompileError(Locale.RES_FUNCTION + ident + Locale.RES_HAS_ONE_PARAM);
                            }
                            expr = new BuiltinFunction(kind.intValue(), params[0]);
                        } else {
                            FunctionCallExpression call = new FunctionCallExpression();
                            call.name = ident;
                            call.parameters = params;
                            expr = call;
                        }
                    }
                } else if (tkn == '{' && ident.equals("plot")) { 
                    Expression list = compileExpression();
                    if (list instanceof ExpressionList) { 
                        expr = new PlotGraphExpression(((ExpressionList)list).exprs);
                    } else { 
                        throw new CompileError(Locale.RES_SYNTAX_ERROR);
                    }
                } else {
                    return new VariableExpression(ident);
                }
            }
            break;
        default:
            throw new CompileError(Locale.RES_SYNTAX_ERROR);
        }
        tkn = scan();
        return expr;
    }

    int scan() {
        char ch = 0;
        int i = pos;
        int n = str.length();
        while (true) { 
            while (i < n && ((ch = str.charAt(i)) == ' ' || ch == '\n')) {
                i += 1;
            }
            if (i < n && ch == '"') { 
                while (++i < n && (ch = str.charAt(i)) != '"');
                i += 1;
            } else { 
                break;
            }
        }
                
        pos = i;
        if (i == n) {
            return -1;
        }
        if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || ch == '_') {
            while (++i < n && (((ch = str.charAt(i)) >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || ch == '_'));
            name = str.substring(pos, i).toLowerCase();
            pos = i;
            return 'a';
        } else if (ch >= '0' && ch <= '9') {
            while (++i < n && (ch = str.charAt(i)) >= '0' && ch <= '9');
            if (ch == '.') {
                while (++i < n && (ch = str.charAt(i)) >= '0' && ch <= '9');
            }
            if (ch == 'e' || ch == 'E') {
                if (i+1 < n && (ch = str.charAt(i+1)) == '+' || ch == '-') {
                    i += 1;
                }
                while (++i < n && (ch = str.charAt(i)) >= '0' && ch <= '9');
            }
            value = new Real(str.substring(pos, i));
            pos = i;
            return '0';
        } else {
            pos = i+1;
            return ch;
        }
    }

    FunctionExpression function;
    String[]           parameters;
    Real               value;
    String             str;
    String             name;
    int                pos;
    int                tkn;
}